# 内存管理 #

实现：陈艺嘉

初赛系统调用完成进度：3 / 3



## 页表 ##

GalaxyOS在初赛阶段采用和xv6一样的页表结构。

在Sv39配置中，一个RISC-V页表在逻辑上是一个由2<sup>27</sup> (134,217,728) 个页表项（Page Table Entry, PTE）组成的数组。每个PTE包含一个44位的物理页号（Physical Page Number, PPN）和一些标志位。分页硬件通过利用39位中的高27位索引到页表中找到一个PTE来转换一个虚拟地址，并计算出一个56位的物理地址，它的前44位来自于PTE中的PPN，而它的后12位则是从原来的虚拟地址复制过来的。图3.1显示了这个过程，在逻辑上可以把页表看成是一个简单的PTE数组（更完整的描述见图3.2）。页表让操作系统控制虚拟地址到物理地址的转换，其粒度为4096（2<sup>12</sup>）字节的对齐块。这样的分块称为页。

在Sv39 RISC-V中，虚拟地址的高25位不用于地址转换。物理地址也有增长的空间：在PTE格式中，物理页号还有 10 位的增长空间。

<img src="img/Figure-3.1.jpg" alt="image1" style="zoom:80%;" />

如图 3.2 所示，实际上转换过程分三步进行。一个页表以三层树的形式存储在物理内存中。树的根部是一个 4096 字节的页表页，它包含 512 个 PTE，这些 PTE 包含树的下一级页表页的物理地址。每一页都包含 512 个 PTE，用于指向下一个页表或物理地址。分页硬件用 27 位中的高 9 位选择根页表页中的 PTE，用中间 9 位选择树中下一级页表页中的 PTE，用低 9 位选择最后的 PTE。

如果转换一个地址所需的三个 PTE 中的任何一个不存在，分页硬件就会引发一个**缺页异常（page-fault exception）**，让内核来处理这个异常。这种三层结构允许页表在处理大范围的虚拟地址没有被映射这种常见情况时，能够忽略整个页表。

每个 PTE 都包含标志位，用于告诉分页硬件相关的虚拟地址被允许怎样使用。`PTE_V` 表示 PTE 是否存在：如果没有设置，对该页的引用会引起异常（即不允许）。`PTE_R` 控制是否允许指令读取该页。`PTE_W` 控制是否允许指令向该页写入。`PTE_X` 控制 CPU 是否可以将页面的内容解释为指令并执行。`PTE_U` 控制是否允许用户态下的指令访问页面；如果不设置 `PTE_U`， 对应 PTE 只能在内核态下使用。图 3.2 显示了这一切的工作原理。标志位和与分页硬件相关的数据结构定义在（`kernel/riscv.h`）中。

<img src="img/Figure-3.2.jpg" alt="image2" style="zoom: 80%;" />

要告诉硬件使用一个页表，内核必须将对应根页表页的物理地址写入 `satp` 寄存器中。每个 CPU 都有自己的 `satp` 寄存器。一个 CPU 将使用自己的 `satp` 所指向的页表来翻译后续指令产生的所有地址。每个 CPU 都有自己的 `satp`，这样不同的 CPU 可以运行不同的进程，每个进程都有自己的页表所描述的私有地址空间。



## 地址空间

### 地址空间布局

GalaxyOS在QEMU模拟机上的地址空间布局如下所示。

```c
#define SBI_BASE            0x80000000

// the kernel expects there to be RAM
// for use by the kernel and user pages
// from physical address 0x80200000 to PHYSTOP.
#ifndef QEMU
#define KERNBASE                0x80020000
#else
#define KERNBASE                0x80200000
#endif

#define PHYSTOP                 0x80600000

// map the trampoline page to the highest address,
// in both user and kernel space.
#define TRAMPOLINE              (MAXVA - PGSIZE)

// mmap space begins here.
#define VUMMAP                  0x70000000
// map kernel stacks beneath the trampoline,
// each surrounded by invalid guard pages.
// #define KSTACK(p)               (TRAMPOLINE - ((p) + 1) * 2 * PGSIZE)
#define VKSTACK                 0x3EC0000000L

// User memory layout.
// Address zero first:
//   text
//   original data and bss
//   fixed-size stack
//   expandable heap
//   ...
//   TRAPFRAME (p->trapframe, used by the trampoline)
//   TRAMPOLINE (the same page as in the kernel)
#define TRAPFRAME               (TRAMPOLINE - PGSIZE)

#define MAXUVA                  SBI_BASE
```

<img src="img/Figure-3.3.jpg" alt="image3" style="zoom: 80%;" />

内核将 trampoline 和 text（可执行程序的代码段）映射为有 `PTE_R` 和 `PTE_X` 权限的页。内核从这些页读取和执行指令。内核映射的其他页会有 `PTE_R` 和 `PTE_W` 权限，以便内核读写这些页面的内存。守护页的映射是无效的（不设置 `PTE_V`）。

### 创建地址空间

大部分用于操作地址空间和页表的代码都在`vm.c`中。核心数据结构是 `pagetable_t`，它实际上是一个指向 RISC-V 根页表页的指针；`pagetable_t`可以是内核页表，也可以是进程的页表。核心函数是 `walk` 和 `mappages`，前者通过虚拟地址得到 PTE，后者将虚拟地址映射到物理地址。以 `kvm` 开头的函数操作内核页表；以 `uvm` 开头的函数操作用户页表；其他函数同时用于这两种页表。`copyout2`和`copyin2`将数据复制到或复制出被作为系统调用参数的用户虚拟地址；它们在 `vm.c` 中，因为它们需要显式转换用户空间的地址，以便找到相应的物理内存。

在机器启动时，在启动序列的靠前部分，`main` 调用 `kvminit`来创建内核页表。这个调用发生在 xv6 在 RISC-V 启用分页之前，所以地址直接指向物理内存。`kvminit` 首先分配一页物理内存来存放根页表页。然后调用 `kvmmap` 将内核所需要的硬件资源映射到物理地址。这些资源包括内核的指令和数据，`KERNBASE` 到 `PHYSTOP`的物理内存，以及实际上是设备的内存范围。

`kvmmap`调用 `mappages`，它将指定范围的虚拟地址映射到一段物理地址。它将范围内地址分割成多页（忽略余数），每次映射一页的起始地址。对于每个要映射的虚拟地址（页的起始地址），`mapages` 调用 `walk` 找到该地址的最后一级 PTE 的地址。然后，它配置 PTE，使其持有相关的物理页号、所需的权限（`PTE_W`、`PTE_X`和/或`PTE_R`），以及`PTE_V`来标记 PTE 为有效。

`walk`（`kernel/vm.c:72`）模仿 RISC-V 分页硬件查找虚拟地址的 PTE（见图 3.2）。`walk` 每次降低 9 位来查找三级页表。它使用每一级的 9 位虚拟地址来查找下一级页表或最后一级（`kernel/vm.c:78`）的 PTE。如果 PTE 无效，那么所需的物理页还没有被分配；如果 `alloc` 参数被设置，`walk` 会分配一个新的页表页，并把它的物理地址放在 PTE 中。它返回树中最低层PTE的地址。

`main` 调用 `kvminithart`来映射内核页表。它将根页表页的物理地址写入寄存器 `satp` 中。在这之后，CPU 将使用内核页表翻译地址。由于内核使用唯一映射，所以指令的虚拟地址将映射到正确的物理内存地址。

`procinit`，它由 `main` 调用，为每个进程分配一个内核栈。它将每个栈映射在 `KSTACK` 生成的虚拟地址上，这就为栈守护页留下了空间。`kvmmap` 将对应的PTE加入到内核页表中，然后调用 `kvminithart` 将内核页表重新加载到 `satp` 中，这样硬件就知道新的 PTE 了。

### 物理内存分配

内核必须在运行时为页表、用户内存、内核堆栈和管道缓冲区分配和释放物理内存。GalaxyOS使用内核地址结束到 `PHYSTOP` 之间的物理内存来进行运行时分配。它每次分配和释放整个4096 字节的页面。它通过保存空闲页链表，来记录哪些页是空闲的。分配包括从链表中删除一页；释放包括将释放的页面添加到空闲页链表中。

分配器在 `kalloc.c`中。分配器的数据结构是一个可供分配的物理内存页的**空闲链表**，每个空闲页的链表元素是一个结构体 `struct run`。分配器从哪里获得内存来存放这个结构体呢？它把每个空闲页的 `run` 结构体存储在空闲页自身里面，因为那里没有其他东西存储。空闲链表由一个**自旋锁**保护。链表和锁被包裹在一个结构体中，以明确锁保护的是结构体中的字段。

`main` 调用 `kinit` 来初始化分配器。`kinit` 初始化空闲页链表，以保存内核地址结束到 `PHYSTOP` 之间的每一页。内核假设机器有 128M 字节的 RAM。`kinit` 通过调用`freerange` 来添加内存到空闲页链表，`freerange` 则对每一页都调用 `kfree`。PTE 只能指向按4096 字节对齐的物理地址（4096 的倍数），因此 `freerange` 使用 `PGROUNDUP` 来确保它只添加对齐的物理地址到空闲链表中。

`kfree`将被释放的内存中的每个字节设置为1。这将使得释放内存后使用内存的代码（使用悬空引用）将会读取垃圾而不是旧的有效内容；希望这将导致这类代码更快地崩溃。然后 `kfree` 将页面预存入释放列表：它将 `pa`（物理地址）转为指向结构体 `run` 的指针，在 `r->next` 中记录空闲链表之前的节点，并将释放列表设为 `r`。`kalloc`移除并返回空闲链表中的第一个元素。

### 进程的地址空间

每个进程都有一个单独的页表，当 xv6 在进程间切换时，也会改变页表。如图 2.3 所示，一个进程的用户内存从虚拟地址 0 开始，可以增长到 `MAXVA`，原则上允许一个进程寻址 256GB 的内存。

当一个进程要求 xv6 提供更多的用户内存时，xv6 首先使用 `kalloc` 来分配物理页，然后将指向新物理页的 PTE 添加到进程的页表中。Xv6 设置这些 PTE 的 `PTE_W`、`PTE_X`、`PTE_R`、`PTE_U` 和 `PTE_V` 标志。大多数进程不使用整个用户地址空间；xv6 将不使用的 PTE 的 `PTE_V` 位保持为清除状态。

我们在这里看到了几个有趣例子，是关于使用页表的。首先，不同的进程页表将用户地址转化为物理内存的不同页，这样每个进程都有私有的用户内存。第二，每个进程都认为自己的内存具有从零开始的连续的虚拟地址，而进程的物理内存可以是不连续的。第三，内核会映射带有 `trampoline` 代码的页到用户地址空间顶端，因此，有一物理内存页在所有地址空间中都会出。

<img src="img/Figure-3.4.jpg" alt="image4" style="zoom:80%;" />



## 系统调用

### 1. sbrk和brk

这两个系统调用用于进程收缩或增长内存，均采用函数`growproc`实现。而`growproc` 调用 `uvmalloc` 或 `uvmdealloc`，取决于 `n` 是正数还是负数。`uvmalloc`通过 `kalloc` 分配物理内存，并使用 `mappages` 将 PTE 添加到用户页表中。 `uvmdealloc` 调用 `uvmunmap`，它使用 `walk` 来查找 PTE 并使用 `kfree` 来释放它们所引用的物理内存。

### 2. mmap和munmap

mmap用于创建内存映射，该映射可以是映射文件的共享内存映射，也可以是映射设备的匿名映射。munmap则用于取消映射。

GalaxyOS创建了一个结构体用于维护内存中的段，段可以是代码段、数据段、堆段、内存映射段或栈段。在初赛阶段，该段只用于描述内存映射段。

```c
enum seg_type { NONE, LOAD, TEXT, DATA, BSS, HEAP, MMAP, STACK };

struct seg {
	int valid;
	enum seg_type type;

	uint64 addr; // segment start address
	uint64 sz; // segment size
	uint64 mmap;
	uint64 f_off;
	int flag;
};
```

`new_seg()`和`locate_seg()`用于创建和定位段。

```c
// Create segment node
int
new_seg(enum seg_type type, uint64 start, uint64 sz, int flag)
{
  #ifdef DEBUG
    //printf("creating new segment.\n");
  #endif
  struct proc *p = myproc();
  int i, j, found = 0;

  for(i = 0; i < NSEG; i++) {
    if(!found && p->segments[i].valid == 0) {
      j = i;
      found = 1;
      #ifdef DEBUG
        //printf("found available segment.\n");
      #endif
      if (start == 0)
        break;
    }
    else if (start != 0 && (start + sz >= p->segments[i].addr) && (start <= p->segments[i].addr + p->segments[i].sz)) {
      printf("Segment overlapping!\n");
      return -1;
    }
  }

  if(found == 1) {
    p->segments[j].valid = 1;
    p->segments[j].type = type;
    if (start)
      p->segments[j].addr = start;
    else
      p->segments[j].addr = p->sz;
    p->segments[j].sz = sz;
    p->segments[j].mmap = NULL;
    p->segments[j].flag = flag;
    
    #ifdef DEBUG
        //printf("returning with %d.\n", j);
    #endif
    return j;
  }
  else
    return -1;
}

int
locate_seg(enum seg_type type, uint64 start, uint sz)
{
  struct proc *p = myproc();

  int i;
  for(i = 0; i < NSEG; i++) {
    if(p->segments[i].valid == 1 && p->segments[i].type == type && p->segments[i].addr == start && p->segments[i].sz == sz)
      return i;
  }

  return -1;
}
```

#### mmap的内存映射

mmap会先调用`new_seg()`，在进程的内存空间中寻找一块区域用于映射内存。起始地址找到后，mmap会以页为单位进行内存分配和`eread`，大小由传入的长度参数决定，向上取整。

munmap则根据传入的地址调用`locate_seg()`，找到映射段后以页为单位释放内存并`ewrite`，大小由传入的长度参数决定，向上取整。



## 遇到的问题

在记录段区域时，一开始想用链表作为数据结构，但是由于内核中没有`malloc`函数，因此采用数组作为数据结构。



## 参考

[XV6-RISCV中文手册](https://github.com/zhenyu-zang/xv6-riscv-book-Chinese/blob/main/Chapter-3.md)

