// Memory map functions

#include "include/types.h"
#include "include/riscv.h"
#include "include/param.h"
#include "include/syscall.h"
#include "include/mm.h"
#include "include/proc.h"
#include "include/vm.h"
#include "include/string.h"
#include "include/kalloc.h"

int
do_munmap(uint64 start, uint64 len)
{
  struct proc *p = myproc();
  int index;

  //printf("doing munmap!\n");

  index = locate_seg(MMAP, start, len);
  if (index == -1)
    return -1;
  else {
    struct dirent *ep = p->ofile[p->segments[index].mmap]->ep;
    uint64 va = start;
    uint64 offset = p->segments[index].f_off;

    //printf("addr: %d\n", va);

    int a;
    for(a = va; a < va + len; a += PGSIZE) {
        uint64 size = a + PGSIZE < len ? PGSIZE : va + len - a;
        
        pte_t *pte = walk(p->pagetable, PGROUNDDOWN(a), 0);

        if((*pte & PTE_V) == 0) {
          //printf("pte not valid.\n");
          return -1;
        }
        if((*pte & PTE_U) == 0) {
          //printf("pte not user.\n");
          return -1;
        }
          
        if(*pte & PTE_W) {       // write back when PTE_D bit is 1
          uint64 pa = PTE2PA(*pte);
          if(!pa) {
            //printf("can't find pa.\n");
            return -1;
          }
          elock(ep);
          ewrite(ep, 0, pa, offset, size);
          eunlock(ep);
        }
        
        if (a != va) {
          vmunmap(p->pagetable, a, 1, 1);
          vmunmap(p->kpagetable, a, 1, 0);
        }


        offset += size;
    }

    //fileclose(p->ofile[p->segments[index].mmap]);

    p->segments[index].mmap = NULL;
    p->segments[index].valid = 0;
  }

  //printf("munmap success.\n");

  return 0;
}

uint64
do_mmap(uint64 start, uint64 len, int prot, int flags, int fd, uint64 off)
{
  #ifdef DEBUG
    //printf("doing mmap.\n");
  #endif
  struct proc *p = myproc();

  int perm = PTE_V;
  
  if(prot & PROT_READ)
    perm |= PTE_R;
  if(prot & PROT_WRITE)
    perm |= PTE_W;
  if(prot & PROT_EXEC)
    perm |= PTE_X;

  // map with file
  if(fd) {
    if (flags & MAP_FIXED) // fixed mapping
    /*
      TODO: you have to munmap exist mapping before fixed mapping
      no need to implement in pre
    */
      return -1;
    else {
      int index;

      index = new_seg(MMAP, start, len, flags);
      if(index != -1) {
        #ifdef DEBUG
          //printf("new segment created.\n");
        #endif
        p->segments[index].f_off = off;
        p->segments[index].mmap = fd;
        
        struct dirent* ep = p->ofile[fd]->ep;
        if(len + off > ep->file_size)
          return -1;
        
        uint64 va = p->segments[index].addr;
        uint64 offset = off;
        char *mem;
        
        int a;
        for(a = va; a < va + len; a += PGSIZE) {
          uint64 size = a + PGSIZE < len ? PGSIZE : va + len - a;

          mem = kalloc();
          if(mem == NULL) {
            uvmdealloc(p->pagetable, p->kpagetable, a, p->sz);
            return -1;
          } 
          memset(mem, 0, PGSIZE);

          if(mappages(p->pagetable, a, PGSIZE, (uint64)mem, perm | PTE_U) != 0) {
            kfree(mem);
            uvmdealloc(p->pagetable, p->kpagetable, a, p->sz);
            return -1;
          }
          if(mappages(p->kpagetable, a, PGSIZE, (uint64)mem, perm) != 0) {
            int npages = (a - p->sz) / PGSIZE;
            vmunmap(p->pagetable, p->sz, npages + 1, 1);   // plus the page allocated above.
            vmunmap(p->kpagetable, p->sz, npages, 0);
            return -1;
          }
          p->sz += PGSIZE;

          uint64 pa = (uint64)walkaddr(p->pagetable, a);
          elock(ep);
          eread(ep, 0, pa, offset, size);
          eunlock(ep);
        
          if (size != PGSIZE)
            memset((void*)(pa + size), 0, PGSIZE - size);

          offset += size;
        }

        //printf("return with %d.\n", p->segments[index].addr);
        filedup(p->ofile[p->segments[index].mmap]);
        return p->segments[index].addr;
      }
      else
        return -1;
    }
  }
  else {
    /*
      TODO: anonymous mapping
      no need to implement in pre
    */ 
    return -1;
  }
}